@@ -1,3 +1,11 @@
+// ============================================================================
+# include <setjmp.h>
+jmp_buf env;
+// longjmp(env, 1);
+// if(setjmp(env)) return List::create();
+// ============================================================================
+
+
 // [[Rcpp::depends(RcppArmadillo)]]
 // [[Rcpp::depends(RcppParallel)]]
 // [[Rcpp::plugins(cpp11)]]
@@ -7,19 +15,22 @@
 # include "h/macros.hpp"
 # include "h/G.hpp"
 # include "h/funs.hpp"
+
+
+
+
 using namespace RcppParallel;
 using namespace Rcpp;
 
 
 // [[Rcpp::export]]
-double testGdensity(
-    NumericVector x, NumericVector mu, NumericVector sigma, double alpha = 1)
+NumericVector testGlogdensity(
+    NumericMatrix X, NumericVector mu, NumericVector sigma, double alpha = 1)
 {
   G<int, double> gaussian;
   gaussian.alpha = alpha;
   gaussian.mu.assign(mu.begin(), mu.end());
   int d = mu.size();
-  // Rcout << "d = " << d << "\n";
   vec<double> triSigma(std::size_t(d) * (d + 1) / 2);
   fullSigmaToTriSigma(&*sigma.begin(), &triSigma.front(), d);
   if(false)
@@ -28,151 +39,54 @@ double testGdensity(
       Rcout << triSigma[i] << ", ";
     Rcout << "\n\n";
   }
-  bool b = gaussian.computeCholUandSqrtOfDet(triSigma);
-  if(!b)
-  {
-    Rcout << "problematic\n";
-    return 0;
-  }
+  gaussian.computeCholUandLogSqrtOfDet(triSigma);
   vec<double> tmp(mu.size());
-  double pi_ = std::pow(2.0 * M_PI, d * (-0.5));
-  return gaussian.densityEval(&x[0], mu.size(), &tmp[0], pi_);
+  // double logPi_ = std::pow(2.0 * M_PI, d * (-0.5));
+  double logPi_ = std::log(2.0 * M_PI) * d * (-0.5);
+  NumericVector rst(X.ncol());
+  for(int i = 0, iend = X.ncol(); i < iend; ++i)
+  {
+    double *x = &X[0] + i * X.nrow();
+    rst[i] = gaussian.logdensityEval(x, d, &tmp[0], logPi_);
+  }
+  return rst;
 }
 
 
-/*
-template<typename indtype, typename valtype>
-void update1G(G<indtype, valtype> &gaussian, valtype &diffS, // valtype eps,
-              indtype d, indtype Xsize, valtype *X,
-              const valtype *pointWeight, valtype *rowSum, valtype *errorS,
-              valtype *buffer, bool updateAlpha = true)
-  // buffer is a temporary container of size Xsize + d * (d + 1) / 2 + d
-{
-  indtype triMatSize = std::size_t(d) * (d + 1) / 2;
-  valtype *&W = buffer; // The first Xsize elements of buffer stores weights
-  valtype sumW = 0;
-  for(indtype i = 0; i < Xsize; ++i)
-  {
-    W[i] = 0;
-    // if(rowSum[i] >= invInf) W[i] = gaussian.ptr[i] / rowSum[i];
-    if(rowSum[i] > 0) W[i] = gaussian.ptr[i] / rowSum[i];
-    W[i] *= pointWeight[i]; // branch prediction will optimize this away
-    sumW += W[i];
-  }
-
-
-  valtype alphaNew = gaussian.alpha;
-  if(updateAlpha) alphaNew = sumW / Xsize;
-
-
-  diffS = 0;
-  diffS += relaErr<valtype> (alphaNew, gaussian.alpha);
-  gaussian.alpha = alphaNew;
-
-
-  for(indtype i = 0; i < Xsize; ++i) W[i] /= sumW;
-
-
-  valtype *oldmu = buffer + Xsize; // buffer[Xsize] -> buffer[Xsize + d] stores the old mean
-  std::copy(gaussian.mu.begin(), gaussian.mu.end(), oldmu);
-
-
-  // Update mean
-  valtype *mean = &gaussian.mu.front();
-  std::fill(mean, mean + d, 0);
-  for(indtype i = 0; i < Xsize; ++i)
-  {
-    valtype *val= X + i * d;
-    for(indtype j = 0; j < d; ++j)
-    {
-      mean[j] += W[i] * val[j];
-    }
-  }
-
-
-  // relative difference
-  diffS += relaErrSum(mean, oldmu, d);
-
-
-  valtype *Sigma = buffer + Xsize; // buffer[Xsize] -> buffer[Xsize + triMatSize]
-  // stores the new covariance (upper tri)
-  std::fill(Sigma, Sigma + triMatSize, 0);
-
-
-  // update sigma
-  valtype *x_mu = Sigma + triMatSize; // buffer[Xsize + triMatSize] stores x - mu, temporary
-  for(indtype i = 0; i < Xsize; ++i)
-  {
-    valtype *val = X + i * d;
-    for(indtype t = 0; t < d; ++t)
-    {
-      x_mu[t] = val[t] - mean[t];
-    }
-
-
-    valtype *s = Sigma;
-    for(indtype p = 0; p < d; ++p)
-    {
-      for(indtype q = 0; q <= p; ++q) // upper triangle
-      {
-        *s += x_mu[p] * x_mu[q] * W[i];
-        ++s;
-      }
-    }
-  }
-
-
-  // bool b = gaussian.computeCholUandSqrtOfDet(Sigma);
-  bool b = gaussian.computeCholUandSqrtOfDet(Sigma);
-  // After this function, Sigma contains the old covariance matrix
-  if(!b)
-  {
-    gaussian.alpha = 0;
-    diffS += triMatSize;
-    return;
-  }
-
-
-  // relative difference
-  valtype *&cholPrevous = Sigma; // Now Sigma stores the old covariance matrix (upper tri)
-  diffS += relaErrSum(&gaussian.cholU[0], cholPrevous, triMatSize);
-}
-*/
 
 
 template<typename indtype, typename valtype>
-void update1G(G<indtype, valtype> &gaussian, valtype &diffS, // valtype eps,
+void update1G(G<indtype, valtype> &gaussian, // valtype eps,
               indtype d, indtype Xsize,
               valtype *X, const valtype *pointWeight,
-              valtype *rowSum, valtype *errorS, valtype *buffer)
+              valtype *rowSum,
+              valtype *buffer,
+              valtype embedNoise)
   // buffer is a temporary container of size Xsize + d * (d + 1) / 2 + d
 {
+
+
   indtype triMatSize = std::size_t(d) * (d + 1) / 2;
   valtype *&W = buffer; // The first Xsize elements of buffer stores weights
   valtype sumW = 0;
   for(indtype i = 0; i < Xsize; ++i)
   {
-    W[i] = 0;
-    if(rowSum[i] > 0) W[i] = gaussian.ptr[i] / rowSum[i];
-    W[i] *= pointWeight[i]; // Branch prediction will optimize this away
+    W[i] = gaussian.ptr[i] / rowSum[i] * pointWeight[i];
     sumW += W[i];
   }
 
 
-  diffS = 0;
-  // std::cout << "gaussian.updateAlpha = " + std::to_string(gaussian.updateAlpha) + ", ";
+  // diffS = 0;
   if(gaussian.updateAlpha)
   {
-    valtype alphaNew = sumW / Xsize;
-    // Rcout << "alphaNew = " << alphaNew << ", ";
-    diffS += relaErr<valtype> (alphaNew, gaussian.alpha);
-    gaussian.alpha = alphaNew;
+    gaussian.alpha = sumW / Xsize;
   }
 
 
   if(gaussian.updateMean or gaussian.updateSigma)
   {
-    for(indtype i = 0; i < Xsize; ++i) W[i] /= sumW;
+    valtype invSumW = 1.0 / sumW;
+    for(indtype i = 0; i < Xsize; ++i) W[i] *= invSumW;
   }
 
 
@@ -191,8 +105,6 @@ void update1G(G<indtype, valtype> &gaussian, valtype &diffS, // valtype eps,
         mean[j] += W[i] * val[j];
       }
     }
-    // Relative difference
-    diffS += relaErrSum(mean, oldmu, d);
   }
 
 
@@ -201,6 +113,7 @@ void update1G(G<indtype, valtype> &gaussian, valtype &diffS, // valtype eps,
     valtype *Sigma = buffer + Xsize; // buffer[Xsize] -> buffer[Xsize + triMatSize]
     // Stores the new covariance (upper tri)
     std::fill(Sigma, Sigma + triMatSize, 0);
+    for(indtype p = 0; p < d; ++p) Sigma[p * (p + 1) / 2 + p] += embedNoise;
     // Update sigma
     valtype *x_mu = Sigma + triMatSize; // buffer[Xsize + triMatSize] stores x - mu, temporary
     for(indtype i = 0; i < Xsize; ++i)
@@ -217,37 +130,21 @@ void update1G(G<indtype, valtype> &gaussian, valtype &diffS, // valtype eps,
         }
       }
     }
-    bool b = gaussian.computeCholUandSqrtOfDet(Sigma);
-    // After this function, Sigma contains the old covariance matrix
-    if(!b)
-    {
-      gaussian.alpha = 0;
-      diffS += triMatSize;
-      return;
-    }
-    // Relative difference
-    valtype *&cholPrevous = Sigma; // Now Sigma stores the old covariance matrix (upper tri)
-    diffS += relaErrSum(&gaussian.cholU[0], cholPrevous, triMatSize);
+    gaussian.computeCholUandLogSqrtOfDet(Sigma);
   }
-
-
 }
 
 
 
 
-
-
 template<typename indtype, typename valtype>
 struct updateParaConventional: public Worker
 {
   indtype d, Xsize, gmodelSize;
-  // valtype eps;
+  valtype embedNoise;
   valtype *X, *pointWeight, *rowSum;
   G<indtype, valtype> *gmodel;
-  valtype *errorS;
   vec<valtype> *tmpCntr;
-  bool updateAlpha;
   dynamicTasking *dT;
 
 
@@ -257,9 +154,8 @@ struct updateParaConventional: public Worker
     {
       std::size_t objI = 0;
       if(!dT->nextTaskID(objI)) break;
-      update1G(gmodel[objI], errorS[objI], // eps,
-               d, Xsize, X, pointWeight,
-               rowSum, errorS, &tmpCntr[st][0]);
+      update1G(gmodel[objI], d, Xsize, X, pointWeight,
+               rowSum, &tmpCntr[st][0], embedNoise);
     }
   }
 
@@ -271,10 +167,11 @@ struct updateParaConventional: public Worker
     valtype *X,
     valtype *pointWeight,
     valtype *rowSum,
-    G<indtype, valtype> *gmodel, valtype *errorS, indtype NofCPU, bool updateAlpha):
-    d(d), Xsize(Xsize), gmodelSize(gmodelSize), // eps(eps),
-    X(X), pointWeight(pointWeight), rowSum(rowSum), gmodel(gmodel), errorS(errorS),
-    updateAlpha(updateAlpha)
+    G<indtype, valtype> *gmodel,
+    indtype NofCPU,
+    valtype embedNoise):
+    d(d), Xsize(Xsize), gmodelSize(gmodelSize), embedNoise(embedNoise),
+    X(X), pointWeight(pointWeight), rowSum(rowSum), gmodel(gmodel)
   {
     dynamicTasking dt(NofCPU, gmodelSize); dT = &dt;
     vec<vec<valtype> > tmp(NofCPU, vec<valtype> (
@@ -325,9 +222,8 @@ List paraGmmFullinit(
     LogicalVector updateAlpha,
     LogicalVector updateMean,
     LogicalVector updateSigma,
-    bool paraConvergeMaxErr,
-    bool loglikehoodConverge,
-    int loglikehoodConvergeBlock)
+    int loglikehoodConvergeBlock,
+    double embedNoise)
 {
   int d = dat.nrow();
   int Xsize = dat.ncol();
@@ -344,7 +240,6 @@ List paraGmmFullinit(
     for(int i = 0, iend = Gvec.size(); i < iend; ++i, muptr += d, sigmaptr += d * d)
     {
       Gvec[i].updateAlpha = updateAlpha[i];
-      // Rcout << "initialization: Gvec[i].updateAlpha = " << Gvec[i].updateAlpha << "\n";
       Gvec[i].updateMean = updateMean[i];
       Gvec[i].updateSigma = updateSigma[i];
 
@@ -355,7 +250,7 @@ List paraGmmFullinit(
       {
         Gvec[i].cholU.resize(std::size_t(d) * (d + 1) / 2);
         fullSigmaToTriSigma(sigmaptr, &Gvec[i].cholU[0], d);
-        Gvec[i].computeCholUandSqrtOfDet(Gvec[i].cholU);
+        Gvec[i].computeCholUandLogSqrtOfDet(Gvec[i].cholU);
       }
     }
 
@@ -371,92 +266,80 @@ List paraGmmFullinit(
 
 
   // EM
-  NumericVector rowSum(Xsize);
+  double previousLL = -1e300;
+  vec<double> rowSum(Xsize), logRowMax(Xsize);
   {
     double *xw = &xweight[0];
     int gaussianParaN = std::size_t(d + 1) * d / 2 + d + 1;
     double endTime = std::clock() + Nthreads * tlimit * CLOCKS_PER_SEC;
     vec<double> auxCntr;
-    vec<double> errorS;
-    double previousLL = -1e300;
     int NllunderEps = 0; // The last `NllunderEps` iterations resulted in loglikelihood less than `eps`.
-    for(int iter = 0, iterEnd = maxit; iter < iterEnd; ++iter)
+
+
+    // for(int iter = 0; iter < maxit; ++iter)
+    for(int iter = 0; ; ++iter)
     {
-      int GvecSize = Gvec.size();
-      cmptDensity<int, double> (d, Xsize, GvecSize, X, &Gvec[0], Nthreads);
-      cmptRowSum<int, double> (Xsize, GvecSize, &Gvec.front(), &rowSum[0], auxCntr, Nthreads);
-      errorS.assign(GvecSize, 0);
+      // int GvecSize = Gvec.size();
+      cmptLogDensity<int, double> (d, Xsize, Gvec.size(), X, &Gvec[0], Nthreads);
+      // Density computation will tell you if some kernel has collapsed: the threshold is
+      // too low so the kernel is virtually singular during computation. At that time,
+      // cmptLogDensity() will set the log determinant to negative infinity.
+      earseCollapsedGau(Gvec);
+      // Rcout << "Gvec.size() = " << Gvec.size() << "\n";
+
+
+      cmptDensityGivenLogDenistyAndRowSum<int, double> (
+          &Gvec.front(), Gvec.size(), &rowSum[0], &logRowMax[0], Xsize, Nthreads);
+
+
+      if(iter >= maxit) break;
+
+
+      int totalParameters = Gvec.size() * gaussianParaN;
 
 
-      int totalParameters = GvecSize * gaussianParaN;
       updateParaConventional<int, double> (
-          d, Xsize, GvecSize, // eps,
-          X, xw, &rowSum[0], &Gvec.front(), &errorS.front(), Nthreads, updateAlpha);
+          d, Xsize, Gvec.size(), X, xw, &rowSum[0], &Gvec.front(), Nthreads, embedNoise);
 
 
       annihilateGinVec(Gvec, annihilationEPS); // Erase components that have small weights.
-      cleanGaussianKernelNotMeetingEigenRatio(Gvec, d, maxEigenRatio, Nthreads);
+
+
       // The function does nothing if maxEigenRatio <= 0.
       // Erase components that have outrageous eigen ratios.
+      cleanGaussianKernelNotMeetingEigenRatio(Gvec, d, maxEigenRatio, Nthreads);
+      // Only delete Gaussian components if embedded noises are 0.
 
 
       int totalParametersAfterCleansing = Gvec.size() * gaussianParaN;
 
 
-      double avgRelativeErr = 0;
       if(totalParametersAfterCleansing == totalParameters)
       {
-        if(!loglikehoodConverge)
-        {
-          if(paraConvergeMaxErr) avgRelativeErr = *std::max_element(errorS.begin(), errorS.end());
-          else avgRelativeErr = std::accumulate(errorS.begin(), errorS.end(), 0.0) / totalParameters;
-        }
-        else
-        {
-          double ll = 0;
-          for(int i = 0, iend = rowSum.size(); i < iend; ++i) ll += std::log(rowSum[i]);
-          if(ll < (((previousLL > 0) * 2 - 1) * eps + 1) * previousLL) ++NllunderEps;
-          else NllunderEps = 0;
-          previousLL = ll;
-        }
+        double ll = 0;
+        for(int i = 0, iend = rowSum.size(); i < iend; ++i)
+          ll += (std::log(rowSum[i]) + logRowMax[i]) * xw[i];
+        if( ll < ( ((previousLL > 0) * 2 - 1) * eps + 1.0  ) * previousLL ) ++NllunderEps;
+        else NllunderEps = 0;
+        previousLL = ll;
       }
       else
       {
         if(verbose)
           Rcout << "Eigenvalue ratios exceed or component weights fall below thresholds, " <<
             "Gaussian components annihilated.\n";
-        avgRelativeErr = eps * 2;
         NllunderEps = 0;
       }
 
 
-      if(!loglikehoodConverge)
-      {
-        if(avgRelativeErr < eps or double(std::clock()) > endTime) break;
-      }
-      else
-      {
-        if(NllunderEps >= loglikehoodConvergeBlock or double(std::clock()) > endTime) break;
-      }
+      if(NllunderEps >= loglikehoodConvergeBlock or double(std::clock()) > endTime) break;
 
 
-      // if(verbose != 0 and (iter + 1) % verbose == 0)
       if(verbose != 0)
       {
-        if(!loglikehoodConverge)
-        {
-          std::string meanMax = "";
-          if(paraConvergeMaxErr) meanMax += ", max absolute relative change in parameters = ";
-          else meanMax += ", mean absolute relative change in parameters = ";
-          Rcout << "iteration = " << iter + 1 << meanMax << avgRelativeErr <<
-              ", number of remaining kernels = " << Gvec.size() << "\n";
-        }
-        else
-        {
-          Rcout << "iteration = " << iter + 1 << ", log-likelihood = " << previousLL <<
-            ", each of the last " << NllunderEps << " iterations yielded less than " <<
-              eps * 100 << "% increase in log-likelihood\n";
-        }
+        Rcout << "iteration = " << iter + 1 << ", log-likelihood = " << previousLL <<
+          ", each of the last " << NllunderEps << " iterations yielded less than " <<
+            eps * 100 << "% increase in log-likelihood\n";
       }
     }
   }
@@ -493,25 +376,29 @@ List paraGmmFullinit(
     }
 
 
-    // if(!updateAlpha and mu.size() < alpha.size())
-    // {
-    //   std::fill(w.begin(), w.end(), 1.0 / mu.size());
-    //   double r = (mu.size() + 0.0) / alpha.size();
-    //   for(int i = 0, iend = rowSum.size(); i < iend; ++i) rowSum[i] *= r;
-    // }
-    // Compute rowSum and w.
     if(true)
     {
       for(int i = 0, iend = w.size(); i < iend; ++i) w[i] = Gvec[i].alpha;
-      // for(int i = 0, iend = w.size(); i < iend; ++i) Rcout << Gvec[i].alpha << ", ";
       double r = 1.0 / std::accumulate(w.begin(), w.end(), 0.0);
       for(int i = 0, iend = w.size(); i < iend; ++i) w[i] *= r;
-      for(int i = 0, iend = rowSum.size(); i < iend; ++i) rowSum[i] *= r;
     }
 
 
-    return List::create(Named("alpha") = w, Named("mu") = mu,
-                        Named("sigma") = sigma, Named("fitted") = rowSum,
+    double ll = 0;
+    for(int i = 0, iend = rowSum.size(); i < iend; ++i) ll += (std::log(rowSum[i]) + logRowMax[i]) * xweight[i];
+    NumericVector fitted(rowSum.size());
+    for(int i = 0, iend = rowSum.size(); i < iend; ++i)
+      fitted[i] = rowSum[i] * std::exp(logRowMax[i]);
+
+
+    int NtotalPara = Gvec.size() * ((d + 1) * d / 2 + d) + Gvec.size() - 1;
+    return List::create(Named("alpha") = w,
+                        Named("mu") = mu,
+                        Named("sigma") = sigma,
+                        Named("LogLikelihood") = ll,
+                        Named("AIC") = -2 * ll + 2 * NtotalPara,
+                        Named("BIC") = -2 * ll + std::log(Xsize + 0.0) * NtotalPara,
+                        Named("fitted") = fitted,
                         Named("clusterMember") = clust);
   }
 }
@@ -538,11 +425,16 @@ List paraGmm(
     LogicalVector updateAlpha,
     LogicalVector updateMean,
     LogicalVector updateSigma,
-    bool paraConvergeMaxErr,
-    bool loglikehoodConverge,
-    int loglikehoodConvergeBlock
+    int convergenceTail,
+    double embedNoise,
+    bool checkInitialization
   )
 {
+
+
+  if(setjmp(env)) return List::create();
+
+
   if(alpha.size() == 0)
   {
     alpha = NumericVector(G, 1.0 / G);
@@ -558,7 +450,7 @@ List paraGmm(
 
 
   // Validate initial covariance matrices and maxEigenRatio.
-  if(true)
+  if(checkInitialization)
   {
     int d = mu.nrow();
     arma::mat tmp(d, d);
@@ -619,7 +511,7 @@ List paraGmm(
     X, alpha, mu, sigma, Xw, maxCore, maxIter,
     eigenRatioLim, convergenceEPS, alphaEPS, tlimit, verbose,
     updateAlpha, updateMean, updateSigma,
-    paraConvergeMaxErr, loglikehoodConverge, loglikehoodConvergeBlock);
+    convergenceTail, embedNoise);
 }
 
 
@@ -651,5 +543,6 @@ List paraGmm(
 
 
 
+
 
 
