@@ -3,25 +3,22 @@
 # include "macros.hpp"
 
 
-// gaussian kernel class
 template<typename indtype, typename valtype>
 struct G
 {
-  bool updateAlpha, updateMean, updateSigma;
-  // valtype piConstMul;
+  bool updateAlpha, updateMean, updateSigma; // valtype piConstMul;
   valtype alpha; // Weight on this kernel.
   valtype sqrtOfDet; // Square root of the covariance matrix's determinant.
+  valtype logSqrtOfDet;
   vec<valtype> mu; // Kernel's center.
   vec<valtype> cholU; // Upper triangle matrix from Cholesky decomposition of the covariance matrix.
   vec<valtype> ptr; // Densities this kernel projects at a set of points in space. Could be an empty vector.
 
 
-  // G(){}
-  // G(indtype d) { piConstMul = std::pow(2.0 * M_PI, d * (-0.5)); alpha = 1; }
-
-
-  inline bool computeCholUandSqrtOfDet(vec<valtype> &sigma);
-  inline bool computeCholUandSqrtOfDet(valtype *sigma);
+  inline void computeCholUandSqrtOfDet(vec<valtype> &sigma);
+  inline void computeCholUandSqrtOfDet(valtype *sigma);
+  inline void computeCholUandLogSqrtOfDet(vec<valtype> &sigma);
+  inline void computeCholUandLogSqrtOfDet(valtype *sigma);
 
 
   inline void shallowCopy(G &x) // Copy parameters of the kernel but not the densities it projects.
@@ -31,6 +28,7 @@ struct G
     updateSigma = x.updateSigma;
     alpha = x.alpha;
     sqrtOfDet = x.sqrtOfDet;
+    logSqrtOfDet = x.logSqrtOfDet;
     mu = x.mu;
     cholU = x.cholU;
   }
@@ -77,14 +75,14 @@ struct G
   // inline valtype densityEval(valtype *x, indtype d, valtype *M, indtype withAlphaWithPi = 3)
   inline valtype densityEval(valtype *x, indtype d, valtype *M, valtype piConstMul)
   {
-    if(sqrtOfDet <= 0) return 0;
+    // if(sqrtOfDet <= 0) return 0;
 
 
     valtype *mean = &*mu.begin();
     valtype *cl = &*cholU.begin();
 
 
-    // mahalanobis distance computation
+    // Mahalanobis distance computation
     valtype *Mst = M;
     valtype *Mbegin = Mst;
     valtype *Mend = M + d;
@@ -93,7 +91,7 @@ struct G
     *Mst = (*x - *mean) / *cl;
     valtype mahanaD = *Mst * *Mst;
     indtype rowLen = 1;
-    for(;;)
+    while(true)
     {
       ++Mst;
       if(Mst >= Mend) break;
@@ -103,31 +101,53 @@ struct G
       ++rowLen;
       valtype numerator = *x - *mean - std::inner_product(Mbegin, Mst, cl, 0.0);
       valtype denominator = *(cl + (Mst - Mbegin));
-      if(denominator == 0.0) *Mst = std::numeric_limits<valtype>::max();
-      // if(numerator + denominator == 0.0) *Mst = 1.0;
-      else *Mst = numerator / denominator;
-      // *Mst = (*x - *mean - std::inner_product(Mbegin, Mst, cl, 0.0)) / *(cl + (Mst - Mbegin));
-      // if(!std::isfinite(*Mst))
-      // {
-      //   Rcpp::Rcout << "*Mst not finite" << ", ";
-      //   Rcpp::Rcout << "fenzi = " << *x - *mean - std::inner_product(Mbegin, Mst, cl, 0.0) << ", ";
-      //   Rcpp::Rcout << "fenmu = " << *(cl + (Mst - Mbegin)) << "\n";
-      //   Rcpp::Rcout << "sqrtOfDet = " << sqrtOfDet << "\n";
-      // }
+      *Mst = numerator / denominator;
       mahanaD += *Mst * *Mst;
     }
 
 
     valtype rst = std::exp(-mahanaD / 2) / sqrtOfDet;
-    // if(!std::isfinite(mahanaD)) Rcpp::Rcout << "rst = " << rst << "\n";
-    // if(!std::isfinite(rst))
-    // {
-    //   Rcpp::Rcout << "-mahanaD / 2 = " << -mahanaD / 2 << ", ";
-    //   Rcpp::Rcout << "std::exp(-mahanaD / 2) = " << std::exp(-mahanaD / 2) << ", ";
-    //   Rcpp::Rcout << "sqrtOfDet = " << sqrtOfDet << "\n";
-    // };
     return rst * alpha * piConstMul;
   }
+
+
+  inline valtype logdensityEval(valtype *x, indtype d, valtype *M, valtype logPiConstMul)
+  {
+    valtype *mean = &*mu.begin();
+    valtype *cl = &*cholU.begin();
+
+
+    // Mahalanobis distance computation
+    valtype *Mst = M;
+    valtype *Mbegin = Mst;
+    valtype *Mend = M + d;
+
+
+    *Mst = (*x - *mean) / *cl;
+    valtype mahanaD = *Mst * *Mst;
+    indtype rowLen = 1;
+    while(true)
+    {
+      ++Mst;
+      if(Mst >= Mend) break;
+      ++x;
+      ++mean;
+      cl += rowLen;
+      ++rowLen;
+      valtype numerator = *x - *mean - std::inner_product(Mbegin, Mst, cl, 0.0);
+      valtype denominator = *(cl + (Mst - Mbegin));
+      if(numerator == 0 and denominator == 0)
+      {
+        logSqrtOfDet = std::numeric_limits<valtype>::lowest();
+        return -1;
+      }
+      *Mst = numerator / denominator;
+      mahanaD += *Mst * *Mst;
+    }
+    return -mahanaD / 2 + std::log(alpha) + logPiConstMul - logSqrtOfDet;
+  }
+
+
 };
 
 
@@ -140,3 +160,5 @@ struct G
 
 
 
+
+
